/*
 * checker.cc: this file is a part of sdf-attribute from the
 *             Transformers project.
 *
 * Copyright (C) 2005  EPITA Research and Development Laboratory.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include <iomanip>
#include "checker.hh"
#include "grammar.hh"
#include "production.hh"

Checker::Checker()
  : inh_dep_error(0),
    syn_dep_error(0),
    multiple_def_error(0),
    cycle_error(0)
{
}

// FIXME: the checker doesn't report error for start symbol inherited dep.
int Checker::check(Grammar& g, bool compat)
{
  Checker c;
  g.forEachProduction(c, &Checker::removeNonCommitedRules);
  g.forEachProduction(c, &Checker::checkProduction);

  g.forEachProduction(c, &Checker::buildDependencyGraph);
  if (!compat)
    {
      c.computeDependencyGraph(g);
      g.forEachProduction(c, &Checker::checkCycle);
    }

  log_2 << "Checker: errors reported : "
	<< c.inh_dep_error << " inherited, "
	<< c.syn_dep_error << " synthesized, "
	<< c.multiple_def_error << " multiple defs, "
	<< c.cycle_error << " cycles."
	<< std::endl;

  log_2 << "Checker: finished..." << std::endl << std::endl;
  return c.inh_dep_error + c.syn_dep_error + c.multiple_def_error
    + c.cycle_error ? 2 : 0;
}

// Remove non-commited rules.
int Checker::removeNonCommitedRules(Production& p)
{
  Production::EvalRuleList::iterator rule_it;

  for (rule_it = p.evalRules.begin(); rule_it != p.evalRules.end(); )
    if (rule_it->isAdded() && !rule_it->isConfirmed())
      rule_it = p.evalRules.erase(rule_it);
    else
      ++rule_it;
  return 0;
}

// This rule is called once for all productions
int Checker::checkProduction(Production& p)
{
  Production::EvalRuleList::iterator rule_it;

  // check inh_dep of its children
  Production::ChildrenNodeList::iterator child_it;
  for (child_it = p.childrenNode.begin(); child_it != p.childrenNode.end(); ++child_it)
    {
      Production* child = *child_it;
      for (rule_it = child->evalRules.begin(); rule_it != child->evalRules.end(); ++rule_it)
	{
	  EvalRule::AttributeList::iterator dep_it;
	  for (dep_it = rule_it->inh_dep.begin(); dep_it != rule_it->inh_dep.end(); ++dep_it)
	    {
	      // if it's a local attribute of this child, or real inh attribute
	      if (!child->has_root_attribute(*dep_it) &&
		  !p.has_attribute(child->name, *dep_it))
		break;
	    }
	  if (dep_it != rule_it->inh_dep.end())
	    {
	      log_1 << std::setw(70) << std::setfill('*') << "*" << std::setfill(' ') << std::endl
		   << "*** Inherited attribute '" << dep_it->getAttrName() << "' missing." << std::endl
		   << std::endl;
	      (*child_it)->printEvalRules(log_1);
	      log_1 << std::endl;
	      p.printEvalRules(log_1);
	      log_1 << std::endl
		   << "In the second production, you should write something like:" << std::endl
		   << "  " <<(*child_it)->name << "." << dep_it->getAttrName() << " := [...]" << std::endl
		   << std::endl;
	      inh_dep_error++;

	    }
	}
    }

  // check its own syn_deps
  Production::EvalRuleList::iterator evalrule_it;
  for (evalrule_it = p.evalRules.begin(); evalrule_it != p.evalRules.end(); ++evalrule_it)
    {
      EvalRule::AttributeList::iterator dep_it;
      for (dep_it = evalrule_it->syn_dep.begin(); dep_it != evalrule_it->syn_dep.end(); ++dep_it)
	{
	  // this attribute can be defined locally
	  if (p.has_attribute(dep_it->getSymbolName(), *dep_it))
	    continue;

	  // all children named 'dep_it->symbol' must set this attribute
	  for (child_it = p.childrenNode.begin(); child_it != p.childrenNode.end(); ++child_it)
	    {
	      if (dep_it->getSymbolName() == (*child_it)->name)
		if (!(*child_it)->has_root_attribute(*dep_it))
		  {
		    log_1 << std::setw(70) << std::setfill('*') << "*" << std::setfill(' ') << std::endl
			 << "*** Synthesized attribute '" << dep_it->getAttrName() << "' missing." << std::endl
			 << std::endl;
		    (*child_it)->printEvalRules(log_1);
		    log_1 << std::endl;
		    p.printEvalRules(log_1);
		    log_1 << std::endl
			 << "In the first production, you should write something like:" << std::endl
			 << "  root." << dep_it->getAttrName() << " := [...]" << std::endl
			 << std::endl;
		    syn_dep_error++;
		  }
	    }
	}
    }

  // check that attribute are defined only once within a production
  for (unsigned i = 0; i < p.evalRules.size(); i++)
    for (unsigned j = i + 1; j < p.evalRules.size(); j++)
      if (p.evalRules[i].left == p.evalRules[j].left)
	{
	  log_1 << std::setw(70) << std::setfill('*') << "*" << std::setfill(' ') << std::endl
	       << "*** Attribute '" << p.evalRules[i].left.getAttrName() << "' is defined multiple times." << std::endl
	       << std::endl;
	  p.printEvalRules(log_1);
	  log_1 << std::endl;
	  multiple_def_error++;
	}
  return 0;
}

// FIXME: should create the graph in live.
int	Checker::buildDependencyGraph(Production& p)
{
  Production::EvalRuleList::iterator rule_it;

  for (rule_it = p.evalRules.begin(); rule_it != p.evalRules.end(); ++rule_it)
    {
      AttrNode& left_an = p.anl_.addAttrNode(&rule_it->left);
      AttrNode* left_dan = NULL;
      if (rule_it->left.isRoot())
	left_dan = &p.dAnl_.addAttrNode(&rule_it->left);

      EvalRule::AttributeList::iterator dep_it;
      for (dep_it = rule_it->inh_dep.begin(); dep_it != rule_it->inh_dep.end(); ++dep_it)
	{
	  AttrNode& an = p.anl_.addAttrNode(&*dep_it);
	  an.addLink(left_an);

	  AttrNode& dan = p.dAnl_.addAttrNode(&*dep_it);
	  // FIXME: should add _all_ paths from Dp
	  if (rule_it->left.isRoot())
	    dan.addLink(*left_dan);
	}

      for (dep_it = rule_it->syn_dep.begin(); dep_it != rule_it->syn_dep.end(); ++dep_it)
	{
	  AttrNode& an = p.anl_.addAttrNode(&*dep_it);
	  an.addLink(left_an);
	}
    }
  return 0;
}

// Algorithm for testing circularity.
// Adapted from Semantics of Context-Free languages, D. Knuth.
int	Checker::computeDependencyGraph(Grammar& g)
{
  Grammar::ProdTableType::iterator it;
  bool has_changed = true;

  log_3 << "Checker::computeDependencyGraph()" << std::endl;
  while (has_changed)
    {
      has_changed = false;
      for (it = g.prod_table.begin(); it != g.prod_table.end(); ++it)
	{
	  Production& p = it->second;
	  Production::ChildrenNodeList::iterator child_it;
	  AttrNodeIter node_it, tmp_it;

	  log_5 << "*** CHECK " << p;

	  // Create a temporary graph, ll, from our partial dep graph.
	  AttrNodeList ll(p.anl_);

	  // Paste Dq_1', ... Dq_np' into ll (Dp[Dq_1', ... Dq_np']).
	  for (child_it = p.childrenNode.begin(); child_it != p.childrenNode.end(); ++child_it)
	    ll.mergeList((*child_it)->dAnl_);

	  log_6 << "ll after merge " << ll;
	  log_6 << "dAnl_ before " << p.dAnl_;

	  for (child_it = p.childrenNode.begin(); child_it != p.childrenNode.end(); ++child_it)
	    for (node_it = ll.begin(); node_it != ll.end(); ++node_it)
	      if (node_it->second.isUp())
		{
		  AttrNode& n = node_it->second;
		  AttrNode& alpha = p.dAnl_.find(n.getUid())->second;
		  std::vector<AttrNode*> node_stack;

		  log_6 << "push attr " << n << std::endl;
		  for (unsigned i = 0; i < n.nbOut_; i++)
		    node_stack.push_back(&ll.find(n.out_[i])->second);

		  // Follow all links starting from alpha, check for link that
		  // end in the list Attribute(p).
		  while (!node_stack.empty())
		    {
		      AttrNode& n = *node_stack.back();
		      node_stack.pop_back();
		      log_6 << "check attr " << n << std::endl;

		      // Found ! Add a link alpha -> alpha_p.
		      if (n.isUp() && p.dAnl_.find(n.getUid()) != p.dAnl_.end())
			{
			  AttrNode& alpha_p = p.dAnl_.find(n.getUid())->second;
			  if (alpha.addLink(alpha_p))
			    {
			      has_changed = true;
			      log_6 << "is a up attr, add link " << alpha_p << std::endl;
			    }
			}
		      else
			for (unsigned i = 0; i < n.nbOut_; i++)
			  {
			    assert(n.out_[i] != n.getUid());
			    node_stack.push_back(&ll.find(n.out_[i])->second);
			  }
		    }
		}
	  log_6 << "dAnl_ after " << p.dAnl_;
	}
    }

  return 0;
}

// Search if there are oriented paths -> there are cycles.
int	Checker::checkCycle(Production& p)
{
  AttrNodeIter node_it;

  log_3 << "\nChecker::checkCyle() for " << p;
  for (node_it = p.dAnl_.begin(); node_it != p.dAnl_.end(); ++node_it)
    if (node_it->second.isUp())
      {
	std::vector<AttrNode*> node_stack;
	AttrNode& alpha = node_it->second;

	log_6 << "push " << alpha << std::endl;
	node_stack.push_back(&alpha);
	while (!node_stack.empty())
	  {
	    AttrNode& tmp = *node_stack.back();
	    node_stack.pop_back();
	    log_6 << "check attr " << tmp << std::endl;
	    for (unsigned i = 0; i < tmp.nbOut_; i++)
	      if (tmp.out_[i] == alpha.getUid())
		{
		  log_1 << std::setw(70) << std::setfill('*') << "*" << std::setfill(' ') << std::endl
			<< "*** Cycle detected on attribute '" << *alpha.getAttr()
			<< "' in the production:" << std::endl
			<< p << std::endl;
		  cycle_error++;
		}
	      else
		node_stack.push_back(&p.dAnl_.find(tmp.out_[i])->second);
	  }
      }
   return 0;
}
