/*
 * evalrule.cc: this file is a part of sdf-attribute from the
 *              Transformers project.
 *
 * Copyright (C) 2005  EPITA Research and Development Laboratory.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include "checker.hh"
#include "evalrule.hh"
#include "production.hh"

std::string EvalRule::cur_name;

EvalRule::EvalRule(ATerm t, Production& p)
  : left(t, p), query(false), added(false), confirmed(false)
{
}

EvalRule::EvalRule(Attribute attr)
  : left(attr), query(false), added(false), confirmed(false)
{
}


/*
** Like a 'collect(attribute(_,_,_))'.
*/
void EvalRule::find_dependencies(ATerm code, Production& prod)
{
  if (ATmatch(code, "attribute(<term>, <term>, <term>)", NULL, NULL, NULL))
    {
      Attribute attr(code, prod);
      addDependency(attr);
    }
  else if (ATgetType(code) == AT_APPL)
    {
      int max = ATgetArity(ATgetAFun(code));
      for (int i = 0; i < max; i++)
	find_dependencies(ATgetArgument(code, i), prod);
    }
  else if (ATgetType(code) == AT_LIST)
    {
      ATermList tl = (ATermList) code;
      while (!ATisEmpty(tl))
	{
	  find_dependencies(ATgetFirst(tl), prod);
	  tl = ATgetNext(tl);
	}
    }
}

void EvalRule::addDependency(const Attribute& dep)
{
  if (dep.isRoot())
    inh_dep.push_back(dep);
  else
    syn_dep.push_back(dep);
}


void EvalRule::confirmDependency(Production& p)
{
  if (confirmed)
    return;

  query = false;
  confirmed = true;

  if (left.isRoot() && syn_dep.size() == 1)
    {
      // root.foo := !NT.bar
      Attribute& right = syn_dep[0];
      aterm = ATmake("term(attribute-rule(" \
		     "None," \
		     "attribute(root,Some(namespace-head(namespace-name(<str>))),<str>)," \
		     "Build(attribute(Sym(<str>),Some(namespace-head(namespace-name(<str>))),<str>))))",
		     left.getNs().c_str(), left.getAttrName().c_str(),
		     right.getLabelName().c_str(), right.getNs().c_str(), right.getAttrName().c_str());
    }
  else if (left.isRoot() && syn_dep.size() > 1)
    {
      // root.foo := <concat> [NT1.bar, NT2.bar, NT3.bar]

      // FIXME: ATerm is certainly wrong.
      ATermList atl = ATmakeList(0);
      for (unsigned i = 0; i < syn_dep.size(); i++)
	{
	  Attribute& right = syn_dep[i];
	  ATerm t = ATmake("attribute(Sym(<str>),Some(namespace-head(namespace-name(<str>))),<str>)" ,
			   right.getLabelName().c_str(), right.getNs().c_str(), right.getAttrName().c_str());
	  atl = ATappend(atl, t);
	}

      aterm = ATmake("term(attribute-rule(" \
		     "None," \
		     "attribute(root,Some(namespace-head(namespace-name(<str>))),<str>)," \
		     "BA(Call(SVar(\"conc\"),[]), Op(\"\", <term>))))",
		     left.getNs().c_str(), left.getAttrName().c_str(),
		     atl);
    }
  else if (!left.isRoot() && inh_dep.size() == 1)
    {
      // NT.foo := !root.bar
      Attribute& right = inh_dep[0];
      aterm = ATmake("term(attribute-rule(" \
		     "None," \
		     "attribute(Sym(<str>),Some(namespace-head(namespace-name(<str>))),<str>)," \
		     "Build(attribute(root,Some(namespace-head(namespace-name(<str>))),<str>))))",
		     left.getLabelName().c_str(), left.getNs().c_str(), left.getAttrName().c_str(),
		     right.getNs().c_str(), right.getAttrName().c_str());
    }
  else if (left.isRoot() && inh_dep.size() == 1)
    {
      // root.foo := root.bar
      Attribute& right = inh_dep[0];
      aterm = ATmake("term(attribute-rule(" \
		     "None," \
		     "attribute(root,Some(namespace-head(namespace-name(<str>))),<str>)," \
		     "Build(attribute(root,Some(namespace-head(namespace-name(<str>))),<str>))))",
		     left.getNs().c_str(), left.getAttrName().c_str(),
		     right.getNs().c_str(), right.getAttrName().c_str());
    }
  else if (left.isRoot() && inh_dep.size() == 0 && syn_dep.size() == 0)
    {
      // root.foo := ![]
      aterm = ATmake("term(attribute-rule(" \
		     "None," \
		     "attribute(root,Some(namespace-head(namespace-name(<str>))),<str>)," \
		     "Build(Op(\"Nil\", []))))",
		     left.getNs().c_str(), left.getAttrName().c_str());
    }
  else
    {
      log_0 << "Error: don't know how to confirm this rule: " << *this << std::endl;
      return;
    }
  ATprotect(&aterm);
  p.attr_list = ATappend(p.attr_list, aterm);
}

const Attribute* EvalRule::getSynDep() const
{
  AttributeList::const_iterator it;
  for (it = syn_dep.begin(); it != syn_dep.end(); ++it)
    if (it->getAttrName().find(cur_name) != std::string::npos)
      return &*it;
  return NULL;
}

const Attribute* EvalRule::getInhDep() const
{
  AttributeList::const_iterator it;
  for (it = inh_dep.begin(); it != inh_dep.end(); ++it)
    if (it->getAttrName().find(cur_name) != std::string::npos)
      return &*it;
  return NULL;
}


bool operator < (const EvalRule& lhs, const EvalRule& rhs)
{
  using namespace std;
  if (lhs.left.isRoot())
    {
      if (rhs.left.isRoot())
	{
// 	  cerr << "syn syn" << endl;
	  return *lhs.getSynDep() <= *rhs.getSynDep();
	}
      else
	{
// 	  cerr << "syn inh : " << endl;
	  if (lhs.getSynDep())
	    return *lhs.getSynDep() < rhs.left;
	  return false;
	}
    }
  else
    {
      if (rhs.left.isRoot())
	{
// 	  cerr << "inh syn" << endl;
	  if (rhs.getSynDep())
	    return lhs.left <= *rhs.getSynDep();
	  return true;
	}
      else
	{
// 	  cerr << "inh inh" << endl;
	  return lhs.left <= rhs.left;
	}
    }
}

std::ostream& operator<< (std::ostream& os, const EvalRule& er)
{
  os << "    " << (er.query ? "<query> " : "") << (er.added ? "<added> " : "") << (er.confirmed ? "<ci> " : "")
     <<  er.left << " := [ ";
  EvalRule::AttributeList::const_iterator it;
  for (it = er.inh_dep.begin(); it != er.inh_dep.end(); ++it)
    os << *it << " ";
  for (it = er.syn_dep.begin(); it != er.syn_dep.end(); ++it)
    os << *it << " ";
  return os << "]" << std::endl;
}
