/*
 * grammar.cc: this file is a part of sdf-attribute from the
 *             Transformers project.
 *
 * Copyright (C) 2005  EPITA Research and Development Laboratory.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include "grammar.hh"

/*
** Fetch all productions of ATerm 't', fill prod_table.
*/
Grammar::Grammar(ATerm t)
{
  other_labels = ATmakeList(0);
  start = extractAterm(t);
  create();
}

Grammar::~Grammar()
{

}

std::string Grammar::getProdName(ATerm elem)
{
  ATerm r;

  if (ATmatch(elem, "cf(sort(<term>))", &r) == ATtrue)
    return get_string(r);
  if (ATmatch(elem, "cf(opt(sort(<term>)))", &r) == ATtrue)
    return get_string(r, "?");
  if (ATmatch(elem, "cf(iter(sort(<term>)))", &r) == ATtrue)
    return get_string(r, "+");
  if (ATmatch(elem, "cf(iter-star(sort(<term>)))", &r) == ATtrue)
    return get_string(r, "*");
  if (ATmatch(elem, "cf(iter-sep(sort(<term>), <term>))", &r, NULL) == ATtrue)
    return get_string(r, "{}");
  if (ATmatch(elem, "cf(iter-star-sep(sort(<term>), <term>))", &r, NULL) == ATtrue)
    return get_string(r, "{}*");
  return std::string("");
}


Production* Grammar::extractAterm(ATerm t)
{
  std::string res;
  ATermList tl;
  ATmatch(t, "parse-table(<int>, <int>, [<list>], <term>, <term>)", NULL, NULL, &tl, NULL, NULL);

  while (!ATisEmpty(tl))
    {
      ATerm e = ATgetFirst(tl);

      /*
      ** If this is a context-free production, add it to our table.
      */
      ATerm prod, sort, attr;
      int iprod;
      ATmatch(e, "label(<term>, <int>)", &prod, &iprod);
      ATmatch(prod, "prod(<term>, <term>, <term>)", NULL, &sort, &attr);
      std::string sort_str = getProdName(sort);

      // Don't retrieve when 'reject' attribute is set.
      bool reject = false;
      ATermList attrs;
      if (ATmatch(attr, "attrs([<list>])", &attrs) == ATtrue)
	{
	  while (!ATisEmpty(attrs))
	    {
	      ATerm e = ATgetFirst(attrs);
	      if (ATmatch(e, "reject") == ATtrue)
		reject = true;
	      attrs = ATgetNext(attrs);
	    }
	}

      if (!reject && sort_str != "")
	{
	  Production new_node;
	  new_node.name = sort_str;
	  new_node.aterm = prod;
	  new_node.iprod = iprod;
	  ATmatch(attr, "attrs([<list>])", &attrs);
	  new_node.attr_list = attrs;
	  ATprotect(&prod);
	  prod_table.insert(std::pair<std::string, Production>(sort_str, new_node));
	}
      else
	{
	  ATprotect(&e);
	  other_labels = ATinsert(other_labels, e);
	}

      /*
      ** Found the start symbol. There should be only one.
      */
      if (ATmatch(prod, "prod([<term>, <term>, <term>], sort(\"<START>\"), <term>)",
		       NULL, &sort, NULL, NULL, NULL, NULL) == ATtrue)
	{
	  res = getProdName(sort);
	  log_3 << "found start symbol: " << res << std::endl;
	}

      tl = ATgetNext(tl);
    }

  log_2 << "found " << prod_table.size() << " productions" << std::endl;
  return &prod_table.find(res)->second;
}


/*
** Retrieve dependencies in evalutation rules.
*/
void Grammar::retrieveRulesDep(Production& n, ATerm attr, ATermList children)
{
  ATermList attr_list, attr_list_copy, label_list = NULL;

  ATmatch(attr, "attrs([<list>])", &attr_list_copy);
  attr_list = attr_list_copy;

  // Try to extract label list.
  while (!ATisEmpty(attr_list))
    {
      ATerm elem = ATgetFirst(attr_list);
      if (ATmatch(elem, "term(labels([<list>]))", &label_list) == ATtrue)
	break;
      attr_list = ATgetNext(attr_list);
    }

  // Sanity check.
  if (label_list == NULL)
    {
      log_0 << "Error: no label for production " << n.name << " ?!?" << std::endl;
      exit(1);
    }

  // Retrieve labels.
  std::vector<std::string> label_indexes;
  while (!ATisEmpty(children))
    {
      ATerm child = ATgetFirst(children);
      std::string child_str = getProdName(child);
      label_indexes.push_back(child_str);
      children = ATgetNext(children);
    }

  // Retrieve non-terminals  for this production and associate their
  // corresponding label.
  while (!ATisEmpty(label_list))
    {
      ATerm elem = ATgetFirst(label_list);
      int index;
      ATerm labelStr;
      ATmatch(elem, "label(<int>, <term>)", &index, &labelStr);

      Production::NT nt;
      nt.index = index;
      nt.label = get_string(labelStr);
      nt.symbol = label_indexes[index - 1];
      n.nonTerminals.push_back(nt);

      label_list = ATgetNext(label_list);
    }

  // Retrieve dependencies in rule code.
  attr_list = attr_list_copy;
  while (!ATisEmpty(attr_list))
    {
      ATerm elem = ATgetFirst(attr_list);
      ATerm left, right;

      if (ATmatch(elem, "term(attribute-rule(<term>, <term>, <term>))",
		  NULL, &left, &right) == ATtrue)
	{
	  ATprotect(&elem);
	  ATprotect(&right);
	  ATprotect(&left);

	  EvalRule& r = n.addRule(Attribute(left, n));
	  r.added = false;
	  r.query = false;
	  r.aterm = elem;
	  r.find_dependencies(right, n);
	}

      attr_list = ATgetNext(attr_list);
    }
}

/*
** Create the dependency graph.
*/
void Grammar::create()
{
  std::vector<Production*> stack;
  stack.reserve(10000);
  stack.push_back(start);

  while (stack.size() > 0)
    {
      Production* n = stack.back();
      stack.pop_back();
      if (n->flag)
	continue;
      n->flag = true;

      ATermList children;
      ATerm attr;
      ATmatch(n->aterm, "prod([<list>], <term>, <term>)", &children, NULL, &attr);

      retrieveRulesDep(*n, attr, children);

      std::pair<ProdTableType::iterator, ProdTableType::iterator> range;
      ProdTableType::iterator it;
      Production::NonTerminalList::iterator nt_it;

      // Retrieve childrenNode and parentsNode for this production.
      for (nt_it = n->nonTerminals.begin(); nt_it != n->nonTerminals.end(); ++nt_it)
	{
	  range = prod_table.equal_range(nt_it->symbol);
	  for (it = range.first; it != range.second; ++it)
	    {
	      n->childrenNode.push_back(&it->second);
	      if (!it->second.flag)
		stack.push_back(&it->second);

	      log_5 << "child " << it->second.name << ", add parent " << n->name << std::endl;
	      it->second.parentsNode.push_back(n);
	    }
	  log_5 << "---" << std::endl;
	}
    }
}


/*
** Recreate an ATerm, similar to the input one, with our modifications.
*/
ATerm	Grammar::rebuildATerm(ATerm at_in)
{
  ATermList outlabels = ATmakeList(0);

  ProdTableType::const_iterator it;
  for (it = prod_table.begin(); it != prod_table.end(); ++it)
    outlabels = ATinsert(outlabels, it->second.getAterm());

  ATermAppl out;
  out = ATsetArgument((ATermAppl) at_in, (ATerm) ATconcat(outlabels, other_labels), 2);
  return (ATerm) out;
}
