/*
 * grammar.hh: this file is a part of sdf-attribute from the
 *             Transformers project.
 *
 * Copyright (C) 2005  EPITA Research and Development Laboratory.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#ifndef GRAMMAR_HH_
# define GRAMMAR_HH_

#include <deque>
#include "checker.hh"
#include "production.hh"

class Grammar
{
public:
  Grammar(ATerm);
  ~Grammar();

  ATerm		rebuildATerm(ATerm in);
  Production*	getStartSymbol();

  template <typename T>
  int forEachProduction(T& obj, int (T::*apply)(Production& p) = &T::operator());
  template <typename T>
  int prefixTraverse(T& obj, int (T::*apply)(Production&));
  template <typename T>
  int suffixTraverse(T& obj, int (T::*apply)(Production&));

  template <typename T, typename U>
  int forEachSameProduction(std::string name, U& ref, T& obj,
			    int (T::*apply)(U& ref, Production& p));

  typedef std::multimap<std::string, Production> ProdTableType;
  ProdTableType	prod_table;

private:
  Production*	start;
  ATermList	other_labels;

  std::string	getProdName(ATerm elem);
  Production*	extractAterm(ATerm t);
  void		retrieveRulesDep(Production& n, ATerm attr, ATermList children);
  void		create();
};

/*
** Apply member function 'apply' to all productions.
*/
template <typename T>
inline int Grammar::forEachProduction(T& obj, int (T::*apply)(Production& p))
{
  ProdTableType::iterator it;
  for (it = prod_table.begin(); it != prod_table.end(); ++it)
    {
      if ((obj.*apply)(it->second))
	return 1;
    }
  return 0;
}

/*
** Apply member function 'apply' of object 'obj' to productions that
** have 'name' as sort name.
*/
template <typename T, typename U>
inline int Grammar::forEachSameProduction(std::string name, U& ref, T& obj,
					int (T::*apply)(U& ref, Production& p))
{
  std::pair<ProdTableType::iterator, ProdTableType::iterator> range;
  ProdTableType::iterator it;

  range = prod_table.equal_range(name);
  for (it = range.first; it != range.second; ++it)
    {
      if ((obj.*apply)(ref, it->second))
	return 1;
    }
  return 0;
}


template <typename T>
inline int Grammar::prefixTraverse(T& obj, int (T::*apply)(Production& p))
{
  ProdTableType::iterator it;
  for (it = prod_table.begin(); it != prod_table.end(); ++it)
    it->second.flag = false;

  std::vector<Production*> stack;
  stack.reserve(1000);
  stack.push_back(start);

  while (stack.size() > 0)
    {
      Production& p = *stack.back();
      stack.pop_back();
      p.flag = true;

      int has_changed = (obj.*apply)(p);

      Production::ChildrenNodeList::reverse_iterator rchild_it;
      for (rchild_it = p.childrenNode.rbegin(); rchild_it != p.childrenNode.rend(); ++rchild_it)
	{
	  if (has_changed)
	    (*rchild_it)->flag = false;
	  if (!(*rchild_it)->flag)
	    stack.push_back(*rchild_it);
	}
    }
  return 0;
}

template <typename T>
inline int Grammar::suffixTraverse(T& obj, int (T::*apply)(Production& p))
{
  std::deque<Production*> queue;

  // Clean flags and put 'terminal' production into the queue.
  ProdTableType::iterator it;
  for (it = prod_table.begin(); it != prod_table.end(); ++it)
    {
      it->second.flag = false;
      if (it->second.childrenNode.size() == 0)
	queue.push_back(&it->second);
    }

  while (queue.size() > 0)
    {
      Production& p = *queue.back();
      queue.pop_back();
      if (p.flag)
	continue;
      p.flag = true;

      int has_changed = (obj.*apply)(p);

      Production::ChildrenNodeList::iterator parent_it;
      for (parent_it = p.parentsNode.begin(); parent_it != p.parentsNode.end(); ++parent_it)
	{
	  if (has_changed)
	    (*parent_it)->flag = false;
	  if (!(*parent_it)->flag)
	    queue.push_front(*parent_it);
	}
    }
  return 0;
}

inline Production*	Grammar::getStartSymbol()
{
  assert(start != NULL);
  return start;
}

#endif /* !GRAMMAR_HH_ */
