/*
 * graph.cc: this file is a part of sdf-attribute from the
 *           Transformers project.
 *
 * Copyright (C) 2005  EPITA Research and Development Laboratory.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include <algorithm>
#include "checker.hh"
#include "graph.hh"

AttrNode::AttrNode(Attribute* a)
  : attr_(a), nbIn_(0), nbOut_(0), mIn_(32), mOut_(32)
{
  in_ = new unsigned[mIn_];
  out_ = new unsigned[mOut_];

  std::string id_str = a->getAttrName() + a->getNs();
  eid_ = hashFct(id_str);
  if (!a->isRoot())
    id_str += a->getSymbolName() + a->getLabelName();
  uid_ = hashFct(id_str);

  isUp_ = a->isRoot();
}

AttrNode::~AttrNode()
{
  // delete[] in_;
  // delete[] out_;
}

bool AttrNode::addLink(AttrNode& to)
{
  // Link to ourselve will let our algorithms loop forever.
  if (to.getUid() == uid_)
    return false;

  // Check that this link doesn't exist yet.
  for (unsigned i = 0; i < nbOut_; i++)
    if (out_[i] == to.uid_)
      return false;

  if (nbOut_ + 1 > mOut_)
    assert(false); // FIXME: realloc
  out_[nbOut_++] = to.uid_;

  if (to.nbIn_ + 1 > to.mIn_)
    assert(false); // FIXME: realloc
  to.in_[to.nbIn_++] = uid_;
  return true;
}


bool AttrNode::removeLink(AttrNode& to)
{
  for (unsigned i = 0; i < nbOut_; i++)
    if (out_[i] == to.uid_)
      {
	out_[i] = out_[nbOut_--];
	for (unsigned j = 0; j < to.nbIn_; j++)
	  if (to.in_[j] == uid_)
	    {
	      to.in_[j] = to.in_[to.nbIn_--];
	      return true;
	    }
      }
  return false;
}

std::ostream& operator<< (std::ostream& os, const AttrNode& a)
{
  os << *a.attr_ << " (uid: " << a.uid_ << " eid: " << a.eid_
     << ", nbIn: " << a.nbIn_ << ", nbOut: " << a.nbOut_ << ")";
  return os;
}


AttrNodeList::AttrNodeList()
{

}

// Doing a block-copy of AttrNodeList is a bad idea, pointer
// in the new list will point to the old list.
// Copy attributes in the new list, then recreate all links.
AttrNodeList::AttrNodeList(const AttrNodeList& r)
  : std::map<unsigned, AttrNode>()
{
  AttrNodeConstIter s_it;
  AttrNodeIter d_it, n_it;

  for (s_it = r.begin(); s_it != r.end(); ++s_it)
    insert(std::pair<unsigned, AttrNode>(s_it->first,
					 AttrNode(s_it->second.getAttr())));

  for (s_it = r.begin(), d_it = begin(); s_it != r.end(); ++s_it, ++d_it)
    {
      const AttrNode& s = s_it->second;
      AttrNode& d = d_it->second;
      for (unsigned i = 0; i < s.nbOut_; i++)
	for (n_it = begin(); n_it != end(); ++n_it)
	  if (n_it->first == s.out_[i])
	    {
	      d.addLink(n_it->second);
	      break;
	    }
    }
}

// Create a new AttrNode from 'attr', and push it in the list.
AttrNode& AttrNodeList::addAttrNode(Attribute* attr)
{
  AttrNode an = AttrNode(attr);

  return insert(std::pair<unsigned, AttrNode>(an.getUid(), an)).first->second;
}

// Merge attributes of 'l' into the current list.
// If l(p) and l(q) exists in the current list and Link(l(p),..., l(q)),
// then add this link in the current graph.
// Otherwise ignore links and attributes.
void	AttrNodeList::mergeList(const AttrNodeList& l)
{
  AttrNodeConstIter s_it;
  AttrNodeIter d_it, d2_it;

  for (s_it = l.begin(); s_it != l.end(); ++s_it)
    for (d_it = begin(); d_it != end(); ++d_it)
      {
	const AttrNode& s = s_it->second;
	AttrNode& d = d_it->second;
	assert(s.isUp());

	if (!d.isUp() && d.getEid() == s.getEid())
	  {
	    std::vector<const AttrNode*> node_stack;
	    for (unsigned i = 0; i < s.nbOut_; i++)
	      node_stack.push_back(&l.find(s.out_[i])->second);
	    while (!node_stack.empty())
	      {
		const AttrNode& n = *node_stack.back();
		node_stack.pop_back();
		bool added = false;
		for (d2_it = begin(); d2_it != end(); ++d2_it)
		  if (!d2_it->second.isUp() && d2_it->second.getEid() == n.getEid())
		    {
		      d.addLink(d2_it->second);
		      added = true;
		      break;
		    }
		if (!added)
		  for (unsigned i = 0; i < n.nbOut_; i++)
		    node_stack.push_back(&l.find(n.out_[i])->second);
	      }
	  }
      }
}

std::ostream& operator<< (std::ostream& os, const AttrNodeList& a)
{
  AttrNodeConstIter it;

  os << "---" << std::endl;
  for (it = a.begin(); it != a.end(); ++it)
    os << it->second << std::endl;
  os << "---" << std::endl;
  return os;
}
