/*
 * main.cc: this file is a part of sdf-attribute from the Transformers project.
 *
 * Copyright (C) 2005  EPITA Research and Development Laboratory.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include <iomanip>
#include <unistd.h>
#include <getopt.h>
#include <cstdio>
#include <ext/stdio_filebuf.h>

#include "checker.hh"
#include "propagator.hh"
#include "grammar.hh"
#include "graphic.hh"

struct s_opt
{
  FILE* in;
  FILE* out;
  bool	out_binary;
  int	verbose;
  bool	compat;
  bool	dot_output;
};

extern char	*optarg;
extern int	optind;

std::ostream log_6(std::cerr.rdbuf());
std::ostream log_5(std::cerr.rdbuf());
std::ostream log_4(std::cerr.rdbuf());
std::ostream log_3(std::cerr.rdbuf());
std::ostream log_2(std::cerr.rdbuf());
std::ostream log_1(std::cerr.rdbuf());
std::ostream log_0(std::cerr.rdbuf());

namespace {
  struct option longopts[] = {
    { "help", no_argument, NULL, 'h' },
    { "verbose", required_argument, NULL, 'v' },
    { "input", required_argument, NULL, 'i' },
    { "output", required_argument, NULL, 'o' },
    { "binary", required_argument, NULL, 'b' },
    { "about", no_argument, NULL, 'a' },
    { "version", no_argument, NULL, 'V' },
    { "compat", no_argument, NULL, 'c'},
    { "dot-output", no_argument, NULL, 'd'},
    { NULL, 0, NULL, 0 }
  };

  void	print_about()
  {
    // FIXME: should write SVN revision number
    printf("attr-defs\n");
    exit(0);
  }

  void	display_help(int exit_code)
  {
    printf("Usage: attr-defs [-hiobvaVd]\n");
    printf("\n");
    printf("Options:\n");
    printf(" -h | --help             Display this help\n");
    printf(" -i f | --input f        Read input from f\n");
    printf(" -o f | --output f       Write output to f\n");
    printf(" -b | --binary           Write binary output\n");
    printf(" -v i | --verbose i      Verbosity level i (default 1)\n");
    printf(" -a | --about            Display information about this program\n");
    printf(" -V | --version          Same as --about\n");
    printf(" -c | --compat           Imit attr-desugar-magic behavior\n");
    printf(" -d | --dot-output       Write some graphical ouput to files\n");
    printf("\n");
    printf("Report bugs to <" PACKAGE_BUGREPORT ">\n");
    exit(exit_code);
  }

  void	opt_init(struct s_opt *opt)
  {
    opt->in = stdin;
    opt->out = stdout;
    opt->out_binary = false;
    opt->verbose = 1;
    opt->compat = false;
    opt->dot_output = false;
  }

  void	opt_fill(int argc, char **argv, struct s_opt *opt)
  {
    int	ch;

    while ((ch = getopt_long(argc, argv, "chi:o:bv:aVd", longopts, NULL)) != -1)
      {
	int i;
	for (i = 0; longopts[i].name && ch != longopts[i].val; i++)
	  ;
	if (longopts[i].name != NULL && ch == '?')
	  display_help(1);

	switch (ch)
	  {
	  case 'h':
	    display_help(0);
	  case 'V':
	  case 'a':
	    print_about();
	  case 'b':
	    opt->out_binary = true;
	    continue;
	  case 'c':
	    opt->compat = true;
	    continue;
	  case 'd':
	    opt->dot_output = true;
	    continue;
	  }
	if (optarg == NULL)
	  display_help(1);
	switch (ch)
	  {
	  case 'i':
	    if ((opt->in = fopen(optarg, "r")) == NULL)
	      {
		fprintf(stderr, "Could not open %s for reading\n", optarg);
		exit(1);
	      }
	    break;
	  case 'o':
	    if ((opt->out = fopen(optarg, "w")) == NULL)
	      {
		fprintf(stderr, "Could not open %s for writing\n", optarg);
		exit(1);
	      }
	    break;
	  case 'v':
	    opt->verbose = atoi(optarg);
	    break;
	  }

      }
  }
  void opt_initlog(int verbose)
  {
    // I'm quite sure that is isn't the right way, but it works fine.
    switch (verbose)
      {
      case 0: log_1.setstate(std::ios::badbit);
      case 1: log_2.setstate(std::ios::badbit);
      case 2: log_3.setstate(std::ios::badbit);
      case 3: log_4.setstate(std::ios::badbit);
      case 4: log_5.setstate(std::ios::badbit);
      case 5: log_6.setstate(std::ios::badbit);
      }
  }
} // !namespace

int main(int argc, char** argv)
{
  ATerm at_bottom, at_in, at_out;
  struct s_opt opt;
  int ret_value = 0;

  // Initialize our program and get command-line options.
  ATinit(argc, argv, &at_bottom);
  opt_init(&opt);
  opt_fill(argc, argv, &opt);
  opt_initlog(opt.verbose);

  // Read and parse input ATerm.
  at_in = ATreadFromFile(opt.in);
  Grammar g(at_in);

  // Do the useful work.
  ret_value = Propagator::propagate(g, opt.compat);
  ret_value += Checker::check(g, opt.compat);

  if (opt.dot_output)
    {
      Graphic html;
      html.makeGraphic(g);
    }

  // Convert our structure back to an ATerm and print it.
  if (opt.verbose > 4)
    ret_value = 0;
  if (ret_value == 0)
    {
      at_out = g.rebuildATerm(at_in);
      if (opt.out_binary)
	ATwriteToBinaryFile(at_out, opt.out);
      else
	ATwriteToTextFile(at_out, opt.out);
      fclose(opt.out);
    }

  // That's all, folks.
  return ret_value;
}
