/*
 * production.cc: this file is a part of sdf-attribute from the
 *                Transformers project.
 *
 * Copyright (C) 2005  EPITA Research and Development Laboratory.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include <iomanip>
#include "checker.hh"
#include "production.hh"

bool Production::has_root_attribute(const Attribute& attr) const
{
  EvalRuleList::const_iterator it;
  for (it = evalRules.begin(); it != evalRules.end(); ++it)
    {
      if (it->left.isRoot()
	  && it->left.getAttrName() == attr.getAttrName()
	  && it->left.getNs() == attr.getNs())
	{
	  return true;
	}
    }
  return false;
}

EvalRule& Production::get_root_attribute(const Attribute& attr)
{
  EvalRuleList::iterator it;
  for (it = evalRules.begin(); it != evalRules.end(); ++it)
    {
      if (it->left.isRoot()
	  && it->left.getAttrName() == attr.getAttrName()
	  && it->left.getNs() == attr.getNs())
	return *it;
    }
  log_0 << "Error: can't retrieve attribute " << attr << std::endl;
  exit(1);
  return evalRules[0];
}

bool Production::has_attribute(const std::string& symbol, const Attribute& attr) const
{
  EvalRuleList::const_iterator it;
  for (it = evalRules.begin(); it != evalRules.end(); ++it)
    {
      if (!it->left.isRoot()
	  && it->left.getSymbolName() == symbol
	  && it->left.getAttrName() == attr.getAttrName()
	  && it->left.getNs() == attr.getNs())
	{
	  return true;
	}
    }
  return false;
}

EvalRule& Production::addRule(const Attribute& copy)
{
  EvalRule er(copy);
  er.setAdded();

  evalRules.push_back(er);
  return evalRules.back();
}

const std::string& Production::labelToSymbol(const std::string label)
{
  std::vector<NT>::iterator it;
  for (it = nonTerminals.begin(); it != nonTerminals.end(); ++it)
    if (it->label == label)
      return it->symbol;

  log_1 << "Error: labelToSymbol: label " << label
       << " not found in " << name << std::endl;
  static std::string empty("");
  return empty;
}

std::vector<std::string> Production::symbolToLabelList(const std::string symbol)
{
  std::vector<std::string> res;
  std::vector<NT>::iterator it;
  for (it = nonTerminals.begin(); it != nonTerminals.end(); ++it)
    if (it->symbol == symbol)
      res.push_back(it->label);

  return res;
}


std::ostream& operator<< (std::ostream& os, const Production& p)
{
  int size = 0;
  Production::NonTerminalList::const_iterator nt_it;
  for (nt_it = p.nonTerminals.begin(); nt_it != p.nonTerminals.end(); ++nt_it)
    {
      size += nt_it->label.size() + nt_it->symbol.size() + 2;
      os << nt_it->label << ":" << nt_it->symbol << " ";
    }

  if (size > 50)
    {
      os << std::endl;
      size = 0;
    }
  os << std::setw(50 - size) << "" << "-> " << p.name << std::endl;
  return os;
}

// Same than operator<<, but formatted for dot output
std::ostream& dotPrint (std::ostream& os, const Production& p)
{
  int size = 0;
  Production::NonTerminalList::const_iterator nt_it;
  for (nt_it = p.nonTerminals.begin(); nt_it != p.nonTerminals.end(); ++nt_it)
    {
      size += nt_it->label.size() + nt_it->symbol.size() + 2;
      os << nt_it->label << ":" << nt_it->symbol << " ";
    }

  os << "-> " << p.name;
  return os;
}

void Production::print() const
{
  std::cerr << *this;
}

void Production::printEvalRules(std::ostream& os) const
{
  using namespace std;

  // print production
  os << *this;

  // print attributes
  os << "  {" << endl;
  EvalRuleList::const_iterator evalrule_it;
  for (evalrule_it = evalRules.begin(); evalrule_it != evalRules.end(); ++evalrule_it)
    os << *evalrule_it;
  os << "  }" << endl;
}

ATerm Production::getAterm() const
{
  ATerm atlist = ATmake("attrs(<term>)", attr_list);
  ATprotect(&atlist);
  ATermAppl prod = ATsetArgument((ATermAppl) aterm, atlist, 2);
  ATerm ret = ATmake("label(<term>, <int>)", prod, iprod);
  ATprotect(&ret);
  return ret;
}
