/*
 * propagator.cc: this file is a part of sdf-attribute from the
 *                Transformers project.
 *
 * Copyright (C) 2005  EPITA Research and Development Laboratory.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include <sstream>
#include <deque>
#include "checker.hh"
#include "grammar.hh"
#include "propagator.hh"
#include "production.hh"

Propagator::Propagator()
  : error_flag(0)
{
}

int Propagator::fillSynQuery(Production& p)
{
  int has_changed = 0;
  Production::EvalRuleList::iterator rule_it;
  EvalRule::AttributeList::iterator dep_it;
  Production::ChildrenNodeList::iterator child_it;

  for (rule_it = p.evalRules.begin(); rule_it != p.evalRules.end(); ++rule_it)
    for (dep_it = rule_it->syn_dep.begin(); dep_it != rule_it->syn_dep.end(); ++dep_it)
      {
	for (child_it = p.childrenNode.begin(); child_it != p.childrenNode.end(); ++child_it)
	  {
	    Production* child = *child_it;
	    if (dep_it->getSymbolName() == child->name
		&& !child->has_root_attribute(*dep_it))
	      if (child->nonTerminals.size() == 0)
		{
		  child->addRule(Attribute(dep_it->getNs(), dep_it->getAttrName()));
		}
	      else
		{
		  Production::NonTerminalList::iterator childchild_it;
		  for (childchild_it = child->nonTerminals.begin(); childchild_it != child->nonTerminals.end(); ++childchild_it)
		    {
		      // add syn dep
		      EvalRule& r = child->addRule(Attribute(dep_it->getNs(), dep_it->getAttrName()));
		      r.addDependency(Attribute(childchild_it->symbol, childchild_it->label,
						dep_it->getNs(), dep_it->getAttrName()));
		      has_changed++;
		    }
		}
	  }
      }
  return has_changed;
}

int Propagator::fillInhQuery(Production& p)
{
  int has_changed = 0;
  Production::ChildrenNodeList::iterator child_it;

  for (child_it = p.childrenNode.begin(); child_it != p.childrenNode.end(); ++child_it)
    {
      Production* child = *child_it;
      int evalrule_max = child->evalRules.size();
      // std::cerr << "*** handle " << p.name << " -> " << child->name << " size: " <<child->rules.size() << std::endl;
      // p.print();
      for (int i = 0; i < evalrule_max; ++i)
	{
	  // std::cerr << "* [" << i << "] new size " << child->rules.size() << std::endl;
	  int dep_max = child->evalRules[i].inh_dep.size();
	  for (int j = 0; j < dep_max; j++)
	    {
	      Attribute& dep = child->evalRules[i].inh_dep[j];
	      if (!p.has_attribute(child->name, dep)
		  && !child->has_root_attribute(dep))
		{
		  // std::cerr << "propagate inh attr " << rule.getAttrName() << " from node "
		  //           << p.name << " to child " << child->name << std::endl;
		  std::vector<std::string> labels = p.symbolToLabelList(child->name);
		  std::vector<std::string>::iterator it;
		  for (it = labels.begin(); it != labels.end(); ++it)
		    {
		      EvalRule& r = p.addRule(Attribute(child->name, *it, dep.getNs(), dep.getAttrName()));
		      r.addDependency(Attribute(dep.getNs(), dep.getAttrName()));
		      has_changed++;
		    }
		}
	    }

	}
    }
  return has_changed;
}

int Propagator::cleanInhQuery(Production& p)
{
  Production::ChildrenNodeList::iterator child_it;
  Production::EvalRuleList::iterator rule_it, rule_it2;
  EvalRule::AttributeList::iterator dep_it;

  for (child_it = p.childrenNode.begin(); child_it != p.childrenNode.end(); ++child_it)
    {
      Production* child = *child_it;
      for (rule_it = child->evalRules.begin(); rule_it != child->evalRules.end(); )
	{
	  if (rule_it->isQuery() && !rule_it->left.isRoot())
	    {
	      // FIXME: don't know what will happen with more that one dep
	      for (dep_it = rule_it->inh_dep.begin(); dep_it != rule_it->inh_dep.end(); ++dep_it)
		{
		  if (dep_it->isRoot())
		    {
		      for (rule_it2 = p.evalRules.begin(); rule_it2 != p.evalRules.end(); ++rule_it2)
			if (!rule_it2->isQuery()
			    && !rule_it2->left.isRoot()
			    && rule_it2->left.getSymbolName() == child->name
			    && rule_it2->left.getAttrName() == dep_it->getAttrName())
			  {
			    rule_it->setQuery(false);
			    goto next_inh_rule;
			  }
		    }
		}
	      // no match: this rule is useless
	      rule_it = child->evalRules.erase(rule_it);
	      continue;
	    }
	next_inh_rule:
	  ++rule_it;
	}
    }
  return 0;
}


int Propagator::cleanSynQuery(Production& p)
{
  int try_again = 0;
  Production::EvalRuleList::iterator rule_it, par_rule_it;
  EvalRule::AttributeList::iterator dep_it;
  Production::ChildrenNodeList::iterator child_it, child_it2;

  for (rule_it = p.evalRules.begin(); rule_it != p.evalRules.end(); )
    {
      // Only interested by syn rules.
      if (!rule_it->left.isRoot())
	goto next_rule;

      if (rule_it->isQuery())
	{
	  for (dep_it = rule_it->syn_dep.begin(); dep_it != rule_it->syn_dep.end(); ++dep_it)
	    for (child_it = p.childrenNode.begin(); child_it != p.childrenNode.end(); ++child_it)
	      if ((*child_it)->has_root_attribute(*dep_it))
		goto next_rule;

	  // No child use it: this rule should be useless.
	  rule_it = p.evalRules.erase(rule_it);
	  continue;
	}

      for (child_it = p.parentsNode.begin(); child_it != p.parentsNode.end(); ++child_it)
	{
	  Production* child = *child_it;

	  for (par_rule_it = child->evalRules.begin(); par_rule_it != child->evalRules.end(); ++par_rule_it)
	    for (dep_it = par_rule_it->syn_dep.begin(); dep_it != par_rule_it->syn_dep.end(); ++dep_it)
	      if (dep_it->getSymbolName() == p.name && dep_it->getAttrName() == rule_it->left.getAttrName())
		{
		  g->forEachSameProduction(p.name, *rule_it, *this, &Propagator::addDefaultSynAttribute);
		  if (par_rule_it->isQuery())
		    try_again = 1;
		  par_rule_it->setQuery(false);
		  goto next_rule;
		}
	}

    next_rule:
      ++rule_it;
    }
  return try_again;
}

int Propagator::addDefaultSynAttribute(EvalRule& ref, Production& p)
{
  Production::EvalRuleList::iterator rule_it;
  bool has_one = false;

  for (rule_it = p.evalRules.begin(); rule_it != p.evalRules.end(); ++rule_it)
    {
      if (rule_it->left == ref.left)
	{
	  has_one = true;
	  rule_it->setQuery(false);
	}
    }

  if (!has_one)
    {
      EvalRule& r = p.addRule(ref.left);
      r.setQuery(false);
    }

  return 0;
}

static void purify_attr_name(std::string& name)
{
  // remove lr_
  name = name.substr(3, 9999);

  // remove _syn or _in, if exist
  if (name.length() < 3)
    return;
  else if (name.substr(name.length() - 3, 3) == "_in")
    name = name.substr(0, name.length() - 3);
  else if (name.substr(name.length() - 4, 4) == "_syn")
    name = name.substr(0, name.length() - 4);
}


namespace std
{
  template <>
  struct less<EvalRule*>
  {
    bool operator()(EvalRule const* p1, EvalRule const* p2)
    {
      if (!p1)
        return true;
      if (!p2)
        return false;
      return *p1 < *p2;
    }
  };
}

int Propagator::doLR(Production& p)
{
  using namespace std;
  Production::ChildrenNodeList::iterator child_it;
  Production::EvalRuleList::iterator rule_it, rule_it2;
  EvalRule::AttributeList::iterator dep_it;

  std::map<std::string, std::vector<EvalRule*> > r;
  for (rule_it = p.evalRules.begin(); rule_it != p.evalRules.end(); ++rule_it)
    if (rule_it->left.getAttrName().substr(0, 3) == "lr_")
      {
	// create a pseudo-graph
	std::string cur_name;
	cur_name = rule_it->left.getAttrName();
	purify_attr_name(cur_name);
	r[cur_name].push_back(&*rule_it);
      }

  // order this graph
  std::map<std::string, std::vector<EvalRule*> >::iterator rit;
  for (rit = r.begin(); rit != r.end(); ++rit)
    {
      rit->second[0]->cur_name = rit->second[0]->left.getAttrName();
      purify_attr_name(rit->second[0]->cur_name);
      // cerr << "will sort:" << rit->second[0]->cur_name << endl;
      // for (std::vector<EvalRule*>::iterator it = rit->second.begin(); it != rit->second.end(); ++it)
      // cerr << **it;
      sort(rit->second.begin(), rit->second.end(), less<EvalRule*>());
//        cerr << "sorted:" << endl;
//        for (std::vector<EvalRule*>::iterator it = rit->second.begin(); it != rit->second.end(); ++it)
//        cerr << **it;

      std::string inh_attr_name;
      for (unsigned i = 0; i < rit->second.size(); i++)
	{
	  const Attribute& left = rit->second[i]->left;
	  const Attribute* syndep = rit->second[i]->getSynDep();
	  const Attribute* inhdep = rit->second[i]->getInhDep();

	  // check that attribute name for inherited is always the same
	  if (!left.isRoot())
	    {
	      if (inh_attr_name != "" && inh_attr_name != left.getAttrName())
		log_0 << "Error: inherited names do not match ! (" << inh_attr_name
		     << " - " << left.getAttrName() << ")" << endl;
	      inh_attr_name = left.getAttrName();
	    }
	  if (inhdep != NULL)
	    {
	      if (inh_attr_name != "" && inh_attr_name != inhdep->getAttrName())
		log_0 << "Error: inherited names do not match ! (" << inh_attr_name
		     << " - " << inhdep->getAttrName() << ")" << endl;
	      inh_attr_name = inhdep->getAttrName();
	    }

	  if (syndep != NULL)
	    {
	      // add default looping, for all children that doesn't have this attribute
	      for (child_it = p.childrenNode.begin(); child_it != p.childrenNode.end(); ++child_it)
		{
		  Production* child = *child_it;
		  if (syndep->getSymbolName() == child->name
		      && !child->has_root_attribute(*syndep))
		    {
		      EvalRule& r = child->addRule(Attribute(syndep->getNs(), syndep->getAttrName()));
		      if (inh_attr_name != "")
			r.addDependency(Attribute(syndep->getNs(), inh_attr_name));
		      r.confirmDependency(*child);
		    }
		}
	    }
	}


      // FIXME: surely doesn't handle all the case
      bool firstSyn = !rit->second[rit->second.size() - 1]->left.isRoot();
      for (unsigned i = rit->second.size() - 2; i != UINT_MAX; i--)
	{
	  // cerr << "i " << i << endl;
	  if (rit->second[i]->left.isRoot() && !rit->second[i + 1]->left.isRoot()
	      && rit->second[i + 1]->getInhDep())
	    {
	      std::stringstream ss;
	      ss << rit->second[i]->left.getAttrName() << (firstSyn ? "" : "_") << i;
	      firstSyn = false;
	      rit->second[i]->left.setAttrName(ss.str());
	      const_cast<Attribute*>(rit->second[i + 1]->getInhDep())->setAttrName(ss.str());
	    }
	  else if (rit->second[i]->left.isRoot() && rit->second[i + 1]->left.isRoot())
	    {
	      // Two synthesized attribute -> concat
	      // FIXME: hacky.
	      rit->second[i]->addDependency(*rit->second[i + 1]->getSynDep());
	      rit->second.erase(rit->second.begin() + i + 1);
	    }
	}

      for (unsigned i = 0; i < rit->second.size(); i++)
	{
	  if (rit->second[i]->isAdded())
	    rit->second[i]->confirmDependency(p);
	}
    }
  return 0;
}


int	Propagator::doTD(Production& p)
{
  Production::EvalRuleList::iterator rule_it;

  for (rule_it = p.evalRules.begin(); rule_it != p.evalRules.end(); ++rule_it)
    {
      if (rule_it->isAdded() && rule_it->left.getAttrName().substr(0, 3) == "td_")
	{
	  if (rule_it->left.isRoot())
	    {
	      log_0 << "Error: td_ attribute can't be synthesized. Use bu_ instead. Production:" << std::endl;
	      log_0 << p << std::endl;
	    }
	  else
	    rule_it->confirmDependency(p);
	    }
    }
  return 0;
}

int	Propagator::doBU(Production& p)
{
  Production::EvalRuleList::iterator rule_it;
  Production::ChildrenNodeList::iterator child_it;

  // fetch all dependencies, and build a single evaluation rule with concat.
  Attribute* left = NULL;
  std::vector<const Attribute*> deps;
  p.evalRules.reserve(p.evalRules.size() * 2);
  for (rule_it = p.evalRules.begin(); rule_it != p.evalRules.end(); ++rule_it)
    {
      if (rule_it->isAdded()
	  && !rule_it->isConfirmed()
	  && rule_it->left.getAttrName().substr(0, 3) == "bu_")
	{
	  if (!rule_it->left.isRoot())
	    {
	      log_0 << "Error: td_ attribute can't be inherited. Use td_ instead. Production:" << std::endl;
	      log_0 << p << std::endl;
	    }
	  else
	    {
	      left = &rule_it->left;
	      if (rule_it->getSynDep() != NULL)
		deps.push_back(rule_it->getSynDep());
	    }
	}
    }
  if (left != NULL)
    {
      EvalRule& r = p.addRule(*left);
      for (unsigned i = 0; i < deps.size(); i++)
	{
	  r.addDependency(*deps[i]);

	  // add default looping, for all children that doesn't have this attribute
	  // FIXME: it should really be elsewhere.
	  for (child_it = p.childrenNode.begin(); child_it != p.childrenNode.end(); ++child_it)
	    {
	      Production* child = *child_it;
	      if (deps[i]->getSymbolName() == child->name
		  && !child->has_root_attribute(*left))
		{
		  EvalRule& r = child->addRule(Attribute(left->getNs(), left->getAttrName()));
		  r.confirmDependency(*child);
		}
	    }
	}

      r.confirmDependency(p);
      // other rules should not be committed, so they will be erased.
    }
  return 0;
}


int Propagator::compatFillQuery(Production& p)
{
  int has_changed = 0;
  Production::EvalRuleList::iterator rule_it;
  EvalRule::AttributeList::iterator dep_it;
  Production::ChildrenNodeList::iterator child_it;

  for (rule_it = p.evalRules.begin(); rule_it != p.evalRules.end(); ++rule_it)
    if (!rule_it->left.isRoot() && rule_it->left.getAttrName().length() > 3
	&& rule_it->left.getAttrName().substr(0, 3) == "lr_")
      for (child_it = p.childrenNode.begin(); child_it != p.childrenNode.end(); ++child_it)
	{
	  Production* child = *child_it;
	  if (rule_it->left.getSymbolName() != child->name)
	    continue;

	  std::string ns = rule_it->left.getNs();
	  std::string attr_inh = rule_it->left.getAttrName();
	  std::string attr_syn = attr_inh.substr(0, attr_inh.length() - 2) + "syn";
	  std::string last_attr = attr_inh;

	  if (child->nonTerminals.size() == 0
	      && !child->has_root_attribute(Attribute(ns, attr_syn)))
	    {
	      // Add a looping rule: root.lr_id_syn := !root.lr_id_in
	      EvalRule& r = child->addRule(Attribute(ns, attr_syn));
	      r.addDependency(Attribute(ns, attr_inh));
	      r.confirmDependency(*child);
	    }
	  else if (!child->has_attribute(child->nonTerminals[0].symbol, rule_it->left)
		   && !child->has_root_attribute(Attribute(ns, attr_syn)))
	    {
	      // Add propag rules for each NT
	      // S.lr_id_in := !root.lr_id_in
	      // root.lr_id_inx := !S.lr_id_syn
	      Production::NonTerminalList::iterator childchild_it;
	      for (childchild_it = child->nonTerminals.begin();
		   childchild_it != child->nonTerminals.end();
		   ++childchild_it)
		{
		  EvalRule& r = child->addRule(Attribute(childchild_it->symbol, childchild_it->label,
							 ns, attr_inh));
		  r.addDependency(Attribute(ns, last_attr));
		  r.confirmDependency(*child);

		  last_attr += "x";
		  if (childchild_it + 1 == child->nonTerminals.end())
		    last_attr = attr_syn;

		  EvalRule& r2 = child->addRule(Attribute(ns, last_attr));
		  r2.addDependency(Attribute(childchild_it->symbol, childchild_it->label,
					     ns, attr_syn));
		  r2.confirmDependency(*child);
		}
	      has_changed++;
	    }
	}
  return has_changed;
}

int Propagator::removeQuery(Production& p)
{
  // remove rules that were not looked up by the cleanSynDep algo.
  // this may be a bug.
  Production::EvalRuleList::iterator rule_it;
  for (rule_it = p.evalRules.begin(); rule_it != p.evalRules.end(); )
    {
      if (rule_it->isQuery())
	rule_it = p.evalRules.erase(rule_it);
      else
	++rule_it;
    }
  return 0;
}

int Propagator::debug(Production& p)
{
  p.printEvalRules(log_1);
  log_1 << std::endl;
  return 0;
}

int Propagator::propagate(Grammar& g, bool compat)
{
  Propagator p;
  p.g = &g;

  if (compat)
    {
      log_3 << "******************** compat pass, for lr_*" << std::endl;
      g.prefixTraverse(p, &Propagator::compatFillQuery);
    }

  log_3 << "******************** first pass" << std::endl;
  g.prefixTraverse(p, &Propagator::fillSynQuery);
  g.suffixTraverse(p, &Propagator::fillInhQuery);

  log_3 << "******************** second pass" << std::endl;
  g.suffixTraverse(p, &Propagator::cleanSynQuery);
  g.prefixTraverse(p, &Propagator::cleanInhQuery);

  g.forEachProduction(p, &Propagator::removeQuery);

  log_3 << "******************** propags" << std::endl;
  g.forEachProduction(p, &Propagator::doTD);
  g.forEachProduction(p, &Propagator::doBU);
  if (!compat)
    g.prefixTraverse(p, &Propagator::doLR);

  log_2 << "Propagator: finished..." << std::endl << std::endl;
  return p.error_flag;
}
